#!/usr/bin/python3

import difflib
import logging
import os

import gi
gi.require_version('Gdk', '3.0')
gi.require_version('Gtk', '3.0')
from gi.repository import GLib, Gtk, Gio

from simplegtd.task import markup_for


class TodoTxt(Gtk.ListStore):

    todofile = None
    last_line_cr = False

    def __init__(self):
        Gtk.ListStore.__init__(self, str, str)
        self.logger = logging.getLogger(self.__class__.__name__)

    def __disestablish_handler(self):
        self.monitor_dir.disconnect(self.monitor_handle)

    def __establish_handler(self):
        self.monitor_handle = self.monitor_dir.connect("changed", self.dir_changed)

    def dir_changed(self, unused_monitor, unused_f, newf, event):
        if event == Gio.FileMonitorEvent.CHANGES_DONE_HINT:
            self.__load()
        elif event == Gio.FileMonitorEvent.RENAMED:
            if newf.get_basename() == os.path.basename(self.todofile):
                self.__load()

    def name(self):
        return self.todofile

    @classmethod
    def from_file(klass, filename):
        self = klass()
        self.todofile = filename
        self.giofile_dir = Gio.File.new_for_path(os.path.dirname(self.todofile))
        self.monitor_dir = self.giofile_dir.monitor(Gio.FileMonitorFlags.WATCH_MOVES, None)
        self.__establish_handler()
        self.__load()
        return self

    def close(self):
        if not self.todofile:
            return
        self.__disestablish_handler()
        self.monitor_dir.cancel()

    def remove_at(self, iter_):
        Gtk.ListStore.remove(self, iter_)
        self.__save()

    def new_at(self, iter_):
        if iter_ is not None:
            path = self.get_path(iter_)
            row = path[0]
        else:
            row = 0
        new_iter = Gtk.ListStore.insert(self, row, ["", ""])
        return new_iter

    def get_at(self, iter_):
        return self.get_value(iter_, 0)

    def remove_many(self, iters):
        for iter_ in iters:
            Gtk.ListStore.remove(self, iter_)
        self.__save()

    def edit(self, iter_, new_text):
        '''Edit task at iter_.  Does nothing if the line does not change.'''
        if self.get_value(iter_, 0) == new_text:
            return
        self.set_value(iter_, 0, new_text)
        self.set_value(iter_, 1, markup_for(new_text))
        self.__save()

    def _load_from_lines(self, new_lines):
        existing_lines = [t[0] for t in self]
        self.last_line_cr = new_lines and new_lines[-1] and new_lines[-1][-1] == "\n"
        def strip_cr(line):
            if line and line[-1] == "\n":
                line = line[:-1]
            return line
        new_lines = [strip_cr(l) for l in new_lines]
        diff = difflib.SequenceMatcher()
        diff.set_seqs(existing_lines, new_lines)
        for op, i1, i2, j1, j2 in diff.get_opcodes():
            if op == 'equal':
                pass
            elif op == 'insert':
                r = 0
                for row in new_lines[j1:j2]:
                    self.logger.debug("Inserting row %s", i1 + r)
                    self.insert(i1 + r, row=[row, markup_for(row)])
                    r = r + 1
            elif op == 'delete':
                for r in reversed(range(i1, i2)):
                    self.logger.debug("Removing row %s", r)
                    self.remove(self[r].iter)
            elif op == 'replace':
                if i2 - i1 == j2 - j1:
                    # The number of rows replaced is the same in old and new.
                    for r in range(i2 - i1):
                        self.logger.debug("Replacing row %s with new row %s", i1 + 1, j1 + r)
                        self.set_value(self[i1 + r].iter, 0, new_lines[j1 + r])
                        self.set_value(self[i1 + r].iter, 1, markup_for(new_lines[j1 + r]))
                elif i2 - i1 > j2 - j1:
                    # The number of rows replaced is larger in old than in new.
                    for r in reversed(range((i2 - i1) - (j2 - j1))):
                        self.logger.debug("Removing row %s", i1 + r)
                        self.remove(self[i1+r].iter)
                    for r in range(j2 - j1):
                        self.logger.debug("Replacing row %s with new row %s", i1 + 1, j1 + r)
                        self.set_value(self[i1 + r].iter, 0, new_lines[j1 + r])
                        self.set_value(self[i1 + r].iter, 1, markup_for(new_lines[j1 + r]))
                else: # i2 - i1 < j2 - j1:
                    # The number of rows replaced is larger in new than in old.
                    for r in range((j2 - j1) - (i2 - i1)):
                        self.logger.debug("Inserting row %s", i1 + r)
                        self.insert(i1 + r)
                    for r in range(j2 - j1):
                        self.logger.debug("Replacing row %s with new row %s", i1 + 1, j1 + r)
                        self.set_value(self[i1 + r].iter, 0, new_lines[j1 + r])
                        self.set_value(self[i1 + r].iter, 1, markup_for(new_lines[j1 + r]))
            else:
                assert 0, "not reached: %s" % op
        return

    def __load(self):
        if not self.todofile:
            return
        try:
            with open(self.todofile, "r") as f:
                lines = f.readlines()
                self._load_from_lines(lines)
        except FileNotFoundError:
            pass
        return self

    def __save(self):
        '''Saves the todo tasks list.'''
        if not self.todofile:
            return
        lines = [row[0] + "\n" for row in self]
        if not self.last_line_cr:
            if lines:
                lines[-1] = lines[-1][:-1]
        text = "".join(lines)
        try:
            with open(self.todofile, "r") as f:
                if text == f.read():
                    return
        except FileNotFoundError:
            pass

        self.__disestablish_handler()
        with open(self.todofile, "w") as f:
            f.write(text)
            f.flush()
        GLib.idle_add(self.__establish_handler)



#!/usr/bin/python3

__version__ = "0.0.10"


import collections
import hashlib
import os

import gi
gi.require_version('Gdk', '3.0')
gi.require_version('Gtk', '3.0')
from gi.repository import GObject, GLib, Gdk, Gtk, Gio

import simplegtd.todotxt
import simplegtd.mainwindow
import simplegtd.resource


def hash_path(filename):
    h = hashlib.md5()
    h.update(filename.encode("utf-8", "ignore"))
    h = h.hexdigest()
    return h


class _SimpleGTDAppState(object):

    todo_txts = None
    __app_state_file = None

    def __init__(self, filename):
        self.__app_state_file = filename
        f = GLib.KeyFile.new()
        try:
            f.load_from_file(filename, GLib.KeyFileFlags.NONE)
            self.todo_txts = f.get_string_list("general", "todo_txts")
        except GLib.Error:
            pass
        return self

    def persist_app_state(self):
        f = GLib.KeyFile.new()
        try:
            f.load_from_file(self.__app_state_file, GLib.KeyFileFlags.NONE)
        except GLib.Error:
            pass
        if self.todo_txts is not None:
            f.set_string_list("general", "todo_txts", self.todo_txts)
        f.save_to_file(self.__app_state_file)


class SimpleGTD(Gtk.Application, _SimpleGTDAppState):

    def __init__(self):
        Gtk.Application.__init__(self)
        self.config_dir = simplegtd.resource.config_dir()
        self.data_dir = simplegtd.resource.data_home()
        for d in self.config_dir, self.data_dir:
            if not os.path.isdir(d):
                os.makedirs(d)

        settings_file = os.path.join(self.config_dir, "app-settings")
        _SimpleGTDAppState.__init__(self, settings_file)
        self.default_todo_txt = os.path.join(self.data_dir, "todo.txt")
        if not self.todo_txts:
            self.todo_txts = [self.default_todo_txt]

        self.connect("activate", lambda _: [self.add_new_todo_window(x) for x in self.todo_txts] and None)
        self.connect("window-removed", self.todo_window_removed)
        self.models_to_windows = {}

    def delegate_open_file_activated(self, requestor):
        choosefile_dialog = Gtk.FileChooserDialog(
            title="Select an existing TODO.TXT file or create a new one",
            parent=requestor,
            action=Gtk.FileChooserAction.SAVE,
            buttons=(
                Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL,
                Gtk.STOCK_OPEN, Gtk.ResponseType.OK,
                "Use default", Gtk.ResponseType.NONE,
            )
        )
        response = choosefile_dialog.run()
        choosefile_dialog.hide()
        if response == Gtk.ResponseType.OK:
            self.add_new_todo_window(choosefile_dialog.get_filename())
        elif response == Gtk.ResponseType.NONE:
            self.add_new_todo_window(self.default_todo_txt)
        choosefile_dialog.destroy()

    def delegate_new_window_activated(self, requestor):
        for data_file, (unused_model, windows) in self.models_to_windows.items():
            if requestor in windows:
                self.add_new_todo_window(data_file)

    def delegate_exit_activated(self, unused_requestor):
        self.quit()

    def add_new_todo_window(self, data_file):
        try:
            if data_file in self.models_to_windows:
                model = self.models_to_windows[data_file][0]
            else:
                model = simplegtd.todotxt.TodoTxt.from_file(data_file)
            window = simplegtd.mainwindow.SimpleGTDMainWindow(model, os.path.join(self.config_dir, "window-state-" + hash_path(data_file)))
            self.add_window(window)
            if data_file not in self.models_to_windows:
                self.models_to_windows[data_file] = (model, [])
            self.models_to_windows[data_file][1].append(window)
            window.connect("open-file-activated", self.delegate_open_file_activated)
            window.connect("new-window-activated", self.delegate_new_window_activated)
            window.connect("exit-activated", self.delegate_exit_activated)
            window.connect("destroy", self.remove_window)
            window.show_all()
        except BaseException:
            Gtk.main_quit()
            raise

    def todo_window_removed(self, unused_ref, window):
        for data_file, (model, windows) in list(self.models_to_windows.items()):
            if window in windows:
                windows.remove(window)
            if not windows:
                model.close()
                del self.models_to_windows[data_file]
        if not self.models_to_windows:
            self.quit()
        else:
            self.todo_txts = list(self.models_to_windows.keys())
            self.persist_app_state()

    def quit(self):
        if self.models_to_windows:
            # First, we disconnect our own todo_window_removed handler,
            # since while quitting we do not want that handler to handle
            # whittling down the window list.
            self.disconnect_by_func(self.todo_window_removed)
            # Now we persist everything.
            self.todo_txts = list(self.models_to_windows.keys())
            self.persist_app_state()
            for model, windows in list(self.models_to_windows.values()):
                for window in windows:
                    window.destroy()
            for model, windows in list(self.models_to_windows.values()):
                model.close()
        Gtk.main_quit()


def main():
    app = SimpleGTD()
    GLib.idle_add(app.run)
    Gtk.main()


if __name__ == "__main__":
    main()



#!/usr/bin/python3

import collections
import logging

import gi
gi.require_version('Gtk', '3.0')
from gi.repository import Gtk

from simplegtd.task import markup_for


class FilterList(Gtk.ListStore):

    def __init__(self, todotxt):
        Gtk.ListStore.__init__(self, str, str, str, str)
        '''Contains filters to be applied to the task list.

        Columns:
           1. plain text string
           2. sortable representation
           3. filter string
           4. markup to render instead of plain text string
        '''
        self.logger = logging.getLogger(self.__class__.__name__)
        self.rows_that_contain_token = collections.defaultdict(list)
        self.tokens_contained_in_row = collections.defaultdict(list)
        self.token_iter_in_filter = {}
        self.append(["All tasks", "", "", "All tasks"])
        todotxt.connect("row-changed", self.todotxt_row_changed)
        todotxt.connect("row-deleted", self.todotxt_row_deleted)
        todotxt.connect("row-inserted", self.todotxt_row_inserted)
        for n, unused_row in enumerate(todotxt):
            self.todotxt_row_inserted(todotxt, Gtk.TreePath.new_from_indices([n]), None)

    def todotxt_row_inserted(self, todotxt, path, unused_it):
        row = path.get_indices()[0]
        text = todotxt[path][0]
        tokens = text.split()
        for tok in tokens:
            tok = tok.strip()
            if tok.startswith("@") or tok.startswith("+"):
                if row not in self.rows_that_contain_token[tok]:
                    self.rows_that_contain_token[tok].append(row)
                if tok not in self.tokens_contained_in_row[row]:
                    self.tokens_contained_in_row[row].append(tok)
                self.logger.debug("New ref count for %s: %s", tok, len(self.rows_that_contain_token[tok]))
                if tok not in self.token_iter_in_filter:
                    sortable = ("1" if tok.startswith("@") else "2") + tok
                    self.token_iter_in_filter[tok] = self.append([tok, sortable, tok, markup_for(tok)])

    def todotxt_row_deleted(self, unused_todotxt, path):
        row = path.get_indices()[0]
        for tok in self.tokens_contained_in_row[row]:
            self.rows_that_contain_token[tok].remove(row)
            self.logger.debug("New ref count for %s: %s", tok, len(self.rows_that_contain_token[tok]))
            if not self.rows_that_contain_token[tok]:
                it = self.token_iter_in_filter[tok]
                self.remove(it)
                del self.token_iter_in_filter[tok]
        del self.tokens_contained_in_row[row]

    def todotxt_row_changed(self, todotxt, path, it):
        self.todotxt_row_deleted(todotxt, path)
        self.todotxt_row_inserted(todotxt, path, it)

    def get_sorted(self):
        '''Wrap the FilterList into a sortable list.'''
        sorted_filters = Gtk.TreeModelSort(model=self)
        sorted_filters.set_sort_column_id(1, Gtk.SortType.ASCENDING)
        return sorted_filters

    def get_filter_string(self, iter_):
        '''Return the filter string pointed to by iter_.'''
        return self.get_value(iter_, 2)



import os

import gi
gi.require_version('Gdk', '3.0')
gi.require_version('Gtk', '3.0')
from gi.repository import GObject, GLib, Gdk, Gtk

import simplegtd.rememberingwindow
import simplegtd.filterlist
import simplegtd.resource
import simplegtd.views


class SimpleGTDMainWindow(Gtk.ApplicationWindow, simplegtd.rememberingwindow.RememberingWindow):

    __gsignals__ = {
        'open-file-activated': (GObject.SIGNAL_RUN_LAST | GObject.SIGNAL_ACTION, None, ()),
        'new-window-activated': (GObject.SIGNAL_RUN_LAST | GObject.SIGNAL_ACTION, None, ()),
        'close-window-activated': (GObject.SIGNAL_RUN_LAST | GObject.SIGNAL_ACTION, None, ()),
        'exit-activated': (GObject.SIGNAL_RUN_LAST | GObject.SIGNAL_ACTION, None, ()),
        'help-activated': (GObject.SIGNAL_RUN_LAST | GObject.SIGNAL_ACTION, None, ()),
        'text-filter-focus-request': (GObject.SIGNAL_RUN_LAST | GObject.SIGNAL_ACTION, None, ()),
    }

    def __init__(self, todotxt, window_state_file):
        self.selection_filters = []
        self.search_filters = []

        Gtk.ApplicationWindow.__init__(self)
        self.set_default_size(800, 600)
        simplegtd.rememberingwindow.RememberingWindow.__init__(self, window_state_file)

        accel_group = Gtk.AccelGroup()
        self.add_accel_group(accel_group)

        self.add_accelerator("close-window-activated", accel_group, ord('w'), Gdk.ModifierType.CONTROL_MASK, 0)
        self.connect("close-window-activated", lambda _: self.destroy())

        header_bar = Gtk.HeaderBar()
        header_bar.set_property('expand', False)
        if todotxt.name():
            n = todotxt.name()
            bn = os.path.basename(n)
            dn = simplegtd.resource.strip_data_home(os.path.dirname(n))
            if dn:
                dn = simplegtd.resource.shorten_path(dn)
                title = "%s (%s)" % (bn, dn)
            else:
                title = "%s" % bn
            header_bar.set_title(title)
            self.set_title(title)
        else:
            header_bar.set_title('Simple GTD (in memory)')
            self.set_title('Simple GTD (in memory)')
        # FIXME: use subtitle to list total and outstanding tasks, filtered and non-filtered
        header_bar.set_subtitle(simplegtd.resource.shorten_path(todotxt.name() or "(no file)"))
        header_bar.set_show_close_button(True)
        self.set_titlebar(header_bar)

        text_filter_entry = Gtk.SearchEntry()
        text_filter_entry.set_placeholder_text("Search tasks...")
        text_filter_entry.connect("search-changed", lambda entry: self.filter_text_changed(entry.get_text()))
        text_filter_entry.connect("stop-search", lambda entry: entry.set_text("") or self.task_view.grab_focus())
        text_filter_entry.connect("activate", lambda unused_entry: self.task_view.grab_focus())
        self.add_accelerator("text-filter-focus-request", accel_group, ord('f'), Gdk.ModifierType.CONTROL_MASK, 0)
        self.connect("text-filter-focus-request", lambda _: text_filter_entry.grab_focus())
        header_bar.pack_start(text_filter_entry)

        exit_button = Gtk.Button.new_from_icon_name("application-exit", Gtk.IconSize.LARGE_TOOLBAR)
        exit_button.connect("clicked", lambda _: self.emit("exit-activated"))
        self.add_accelerator("exit-activated", accel_group, ord('q'), Gdk.ModifierType.CONTROL_MASK, 0)
        header_bar.pack_end(exit_button)

        new_view_button = Gtk.Button.new_from_icon_name("window-new", Gtk.IconSize.LARGE_TOOLBAR)
        new_view_button.connect("clicked", lambda _: self.emit("new-window-activated"))
        self.add_accelerator("new-window-activated", accel_group, ord('n'), Gdk.ModifierType.CONTROL_MASK, 0)
        header_bar.pack_end(new_view_button)

        choosefile_button = Gtk.Button.new_from_icon_name("document-open", Gtk.IconSize.LARGE_TOOLBAR)
        choosefile_button.connect("clicked", lambda _: self.emit("open-file-activated"))
        self.add_accelerator("open-file-activated", accel_group, ord('o'), Gdk.ModifierType.CONTROL_MASK, 0)
        header_bar.pack_end(choosefile_button)

        help_button = Gtk.Button.new_from_icon_name("system-help", Gtk.IconSize.LARGE_TOOLBAR)
        help_button.connect("clicked", lambda _: self.show_shortcuts_window())
        self.add_accelerator("help-activated", accel_group, Gdk.KEY_F1, 0, 0)
        self.connect("help-activated", lambda _: self.show_shortcuts_window())
        header_bar.pack_end(help_button)

        self.task_view = simplegtd.views.TaskView()
        task_view_scroller = Gtk.ScrolledWindow()
        task_view_scroller.set_policy(Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.AUTOMATIC)
        task_view_scroller.add(self.task_view)

        self.filter_view = simplegtd.views.FilterView()
        self.filter_view.get_selection().connect("changed", self.filter_selection_changed)
        filter_view_scroller = Gtk.ScrolledWindow()
        filter_view_scroller.set_policy(Gtk.PolicyType.NEVER, Gtk.PolicyType.AUTOMATIC)
        filter_view_scroller.add(self.filter_view)

        filters = simplegtd.filterlist.FilterList(todotxt)
        self.task_view.set_model(todotxt)
        self.filter_view.set_model(filters)

        GLib.idle_add(self.task_view.focus_first)
        GLib.idle_add(self.filter_view.select_first)

        paned = Gtk.Paned()
        paned.set_wide_handle(True)
        paned.pack1(filter_view_scroller, False, False)
        paned.add2(task_view_scroller)

        self.add(paned)
        self.task_view.grab_focus()

    def filter_selection_changed(self, tree_selection):
        filter_strings = self.filter_view.get_filters_from_selection(tree_selection)
        self.selection_filters = filter_strings
        self.task_view.set_filters(self.selection_filters + self.search_filters)

    def filter_text_changed(self, new_text):
        filter_strings = []
        if new_text.strip():
            filter_strings.append(new_text.strip())
        self.search_filters = filter_strings
        self.task_view.set_filters(self.selection_filters + self.search_filters)

    def show_shortcuts_window(self):
        builder = Gtk.Builder()
        builder.add_from_file(simplegtd.resource.find_data_file("shortcuts-window.ui"))
        shortcuts_window = builder.get_object("shortcuts-simplegtd")
        shortcuts_window.set_transient_for(self)
        shortcuts_window.show_all()
        shortcuts_window.set_property("view-name", "")
        shortcuts_window.set_property("section-name", "shortcuts")



#!/usr/bin/python3

import unittest
from simplegtd.todotxt import TodoTxt

class TodoTxtTest(unittest.TestCase):

    def test_load_from_lines(self):
        t = TodoTxt()
        t._load_from_lines(['a\n', 'b\n', 'c'])
        assert t[0][0] == "a"
        assert t[1][0] == "b"
        assert t[2][0] == "c"
        assert t.last_line_cr is False

    def test_reload_from_lines(self):
        t = TodoTxt()
        t._load_from_lines(['a\n', 'b\n', 'c'])
        t._load_from_lines(['b\n', 'c'])
        assert t[0][0] == "b"
        assert t[1][0] == "c"
        assert len(t) == 2
        assert t.last_line_cr is False

    def test_replace_from_lines(self):
        t = TodoTxt()
        t._load_from_lines(['a\n', 'b\n', 'c'])
        t._load_from_lines(['z\n', 'y\n', 'b\n', 'c'])
        assert t[0][0] == "z", t[0][0]
        assert t[1][0] == "y"
        assert t[2][0] == "b"
        assert t[3][0] == "c"
        assert len(t) == 4
        assert t.last_line_cr is False
        t._load_from_lines(['a\n', 'b\n', 'c'])
        assert t[0][0] == "a", t[0][0]
        assert t[1][0] == "b"
        assert t[2][0] == "c"
        assert len(t) == 3
        t._load_from_lines(['a\n', 'y\n', 'c'])
        assert t[0][0] == "a", t[0][0]
        assert t[1][0] == "y"
        assert t[2][0] == "c"
        assert len(t) == 3

    def test_markup(self):
        t = TodoTxt()
        t._load_from_lines(['2018-09-09 a\n'])
        assert t[0][0] == "2018-09-09 a", t[0][0]
        assert t[0][1] == "<span foreground='green'>2018-09-09</span> a", t[0][1]
        assert t.last_line_cr is True



import os

import xdg.BaseDirectory


def find_data_file(filename):
    '''Raises KeyError if filename is not found in XDG_DATA_DIRS.'''
    for path in (
        [os.path.join(os.path.dirname(__file__), os.path.pardir, os.path.pardir, "data")]
        + [os.path.join(x, "simplegtd") for x in xdg.BaseDirectory.xdg_data_dirs]
    ):
        f = os.path.join(path, os.path.basename(filename))
        if os.path.isfile(f):
            return f
    raise KeyError(f)


def config_dir():
    return os.path.join(xdg.BaseDirectory.xdg_config_home, "simplegtd")


def data_home():
    return os.path.join(xdg.BaseDirectory.xdg_data_home, "simplegtd")


def strip_data_home(path):
    if path.startswith(data_home() + "/") or path == data_home():
        return path[len(data_home()) + 1:]
    return path


def shorten_path(filename):
    if filename.startswith(os.path.expanduser("~/")):
        filename = "~/" + filename[len(os.path.expanduser("~/")):]
    return filename
